import com.cycling74.max.*;

public class RhhRuleSet extends MaxObject
{
  public RhhRuleSet(String[][] rule, int[] stops, int end_a, int end_b, int contin, int start)
  {
    // Specify how many of each type of rule you want: maybe these
    // should be parameters read or passed in from somewhere so you
    // can change them w/o having to recompile?
    
    // This is the max number of rules you are willing to generate
    // before giving up on this grammar.
    int maxRules = 10000;
    // Some assorted integer counters.
    int i = 0;
    int j = 0;

    

    ecsloop:
    while ((end_a > 0) || (end_b > 0) || (contin > 0) || (start > 0))
	{
	    RhhRule genrule = new RhhRule ();

	    // Generate end_a rules, which occupy positions 0 through
	    // stops[0] in rule array.
	    end_a_loop:
	    while (end_a > 0)
		{
		    genrule.endRuleA (rule, (stops[0]-end_a));
		    maxRules--;
		    // Check to see if you failed to generate a legal
		    // grammar in maxRules attempts and give up.
		    if (maxRules == 0)
			System.exit(-1);

		    for (i = 0; i < (stops[0]-end_a); i++)
			if (genrule.isEqual (rule, i, (stops[0]-end_a)) == 1)
			    {
				continue end_a_loop;
			    }
		    // Accept the rule.
		    end_a--;
		}

	    // Generate end_b rules, which occupy positions stops[0] through
	    // stops[1] in rule array.
	    end_b_loop:
	    while (end_b > 0)
		{
		    genrule.endRuleB (rule, (stops[1]-end_b));
		    maxRules--;
		    // Check to see if you failed to generate a legal
		    // grammar in maxRules attempts and give up.
		    if (maxRules == 0)
			System.exit(-1);

		    for (i = stops[0]; i < (stops[1]-end_b); i++)
			if (genrule.isEqual (rule, i, (stops[1]-end_b)) == 1)
			    {
				continue end_b_loop;
			    }
		    // Accept the rule.
		    end_b--;
		}

	    // Generate contin rules, which occupy positions stops[1] through
	    // stops[2] in rule array.
	    contin_loop:
	    while (contin > 0)
		{
		    genrule.conRule (rule, (stops[2]-contin));
		    maxRules--;
		    // Check to see if you failed to generate a legal
		    // grammar in maxRules attempts and give up.
		    if (maxRules == 0)
			System.exit(-1);

		    for (i = stops[1]; i < (stops[2]-contin); i++)
			    if (genrule.isEqual (rule, i, (stops[2]-contin)) == 1)
				{
			    continue contin_loop;
				}
		    // Accept the rule.
		    contin--;
		}

	    // Generate start rules, which occupy positions stops[2] through
	    // stops[3] in rule array.
	    start_loop:
	    while (start > 0)
		{
		    genrule.startRule (rule, (stops[3]-start));
		    maxRules--;
		    // Check to see if you failed to generate a legal
		    // grammar in maxRules attempts and give up.
		    if (maxRules == 0)
			System.exit(-1);

		    for (i = stops[2]; i < (stops[3]-start); i++)
			if (genrule.isEqual (rule, i, (stops[3]-start)) == 1)
			    {
				continue start_loop;
			    }
		    // Accept the rule.
		    start--;
		}

	    // At this point, you have a full rule set.

	    // Test to make sure the rules you have generated don't
	    // have any orphan consequences; that is, everything on
	    // the right side of a contin or start rule has a matching
	    // head on another rule.
	    
	    // Check only contin and start rules. So we need to make
	    // sure that rule[i][3] for contin rules (between stops[1]
	    // and stops[2]) and rule[i][3] for start rules (between
	    // stops[2] and stops[3]) have a match in any rule[i][0]
	    // somewhere other than themselves.
	    pathexistloop:
	    for (i = stops[1]; i < stops[3]; i++)
	      {
		  // Check rule i, which is known to be a start or
		  // contin rule, against all other rules.
		for (j = 0; j < stops[3]; j++)
		  {
		      // Don't check against yourself.
		      if (i == j)
			  break;
		      if (genrule.existPath (rule, j, i) == 1)
			  // Success; rule i has a continuation by
			  // matching header of rule j.
			  continue pathexistloop;
		  }

		// If you get to this point, rule[i] does not have a
		// continuation. The safe remedy is to delete rule[i]
		// and generate a new rule[i].
		if (i < stops[2]) 
		    {
			// A contin rule failed: force generation of a
			// new contin rule.
			contin++;
			rule[i][0] = rule[stops[2]-contin][0];
			rule[i][1] = rule[stops[2]-contin][1];
			rule[i][2] = rule[stops[2]-contin][2];
			rule[i][3] = rule[stops[2]-contin][3];

			// Note: You could, at this point, elect to
			// break pathexistloop as soon as a single
			// rule is rejected. This would mean each
			// rejected rule is immediately regenerated
			// and rechecked. If you don't break, then you
			// will reject perhaps multiple rules and
			// regenerate multiple rules. I don't know
			// which approach is more reasonable. If you
			// reject and generate immediately, you may
			// not have to reject additional rules. On the
			// other hand, it may be more efficient to
			// reject a bunch and regenerate a bunch.

			// break pathexistloop:
		    }
		else
		    {
			// A start rule failed: force generation of a
			// new start rule.
			start++;
			rule[i][0] = rule[stops[3]-start][0];
			rule[i][1] = rule[stops[3]-start][1];
			rule[i][2] = rule[stops[3]-start][2];
			rule[i][3] = rule[stops[3]-start][3];

			// Note: You could, at this point, elect to
			// break pathexistloop as soon as a single
			// rule is rejected. This would mean each
			// rejected rule is immediately regenerated
			// and rechecked. If you don't break, then you
			// will reject perhaps multiple rules and
			// regenerate multiple rules. I don't know
			// which approach is more reasonable. If you
			// reject and generate immediately, you may
			// not have to reject additional rules. On the
			// other hand, it may be more efficient to
			// reject a bunch and regenerate a bunch.

			// break pathexistloop:
		    }
	      }
	  }
    // Finished. Have a complete set of rules with not orphans on the
    // right side!
    
  }
  
  public void printRule(String[][] rule, int[] stops)
  {
  	int i;
  	for (i = 0; i < stops[3]; i++)
		post("Rule" + i + ": " + rule[i][0] + " -> " +
			    rule[i][1] + " " + rule[i][2] + " " + rule[i][3]);
  }
  
  
}
